{****************************************************************************

                     Copyright (c) 1993,94 by Florian Klmpfl

 ****************************************************************************}

unit tree;

  interface

    uses
       globals,symtable,cobjects,asmgen;

    type
       ttreetyp = (addn,muln,subn,divn,
                   modn,assignn,loadn,rangen,
                   ltn,lten,gtn,gten,
                   equaln,unequaln,inn,orn,
                   xorn,shrn,shln,slashn,
                   andn,subscriptn,derefn,addrn,
                   ordconstn,typeconvn,calln,callparan,
                   realconstn,umminusn,asmn,vecn,
                   stringconstn,funcretn,selfn,
                   notn,inlinen,niln,errorn);

       tconverttype = (tc_equal,tc_not_possible,tc_u8bit_2_s32bit,
                       tc_only_rangechecks32bit,tc_s8bit_2_s32bit,
                       tc_u16bit_2_s32bit,tc_s16bit_2_s32bit,
                       tc_s32bit_2_s16bit,tc_s32bit_2_u8bit,
                       tc_s32bit_2_u16bit,tc_string_to_string,
                       tc_cstring_charpointer,tc_cstring_chararray,
                       tc_array_to_pointer,tc_pointer_to_array,
                       tc_char_to_string,tc_u8bit_2_s16bit,
                       tc_u8bit_2_u16bit,tc_s8bit_2_s16bit,
                       tc_s16bit_2_s8bit,tc_s16bit_2_u8bit,
                       tc_u16bit_2_s8bit,tc_u16bit_2_u8bit,
                       tc_s8bit_2_u16bit,tc_s32bit_2_s8bit,
                       tc_int_2_real,tc_cchar_chararray);

       { gibt an, welche Nachfolger eines Knotens }
       { gelscht werden mssen }
       tdisposetyp = (dt_nothing,dt_leftright,dt_left,dt_leftmbright,
                      dt_mbleft,dt_ltree,dt_string,dt_typeconv,dt_inlinen,
                      dt_mbleft_and_method);

       plocation = ^tlocation;

       { Angabe ber den Ort eines Operanden im Speicher }
       { LOC_FPU_STACK   auf dem FPU-Stack }
       { LOC_REGISTER    in Prozessorregister }
       { LOC_MEM         im Speicher
       { LOC_REFERENZ    wie LOC_MEM, nur gleichzeitig lvalue }
       { LOC_JUMP        nur bool'sche Resultate, Sprung zu false- oder }
       {                 truelabel }
       { LOC_FLAGS       nur bool'sche Rsultate, Flags sind gesetzt }

       tloc = (LOC_FPUSTACK,LOC_REGISTER,LOC_MEM,LOC_REFERENZ,LOC_JUMP,
               LOC_FLAGS);

       tresflags = (F_E,F_NE,F_G,F_L,F_GE,F_LE);

       tlocation = record
          case loc : tloc of
             LOC_REGISTER : (register : tregister);
             LOC_MEM,LOC_REFERENZ : (referenz : treferenz);
             LOC_FPUSTACK : ();
             LOC_JUMP : ();
             LOC_FLAGS : (resflags : tresflags);
       end;

       ptree = ^ttree;

       ttree = record
          resulttype : pdef;
          error : boolean;
          registers32,registersfpu : integer;
          disposetyp : tdisposetyp;
          swaped : boolean; { wird auf true gesetzt, wenn linker und }
                            { rechter Operand vertauscht sind        }
          location : tlocation;
          left,right : ptree;
          case treetype : ttreetyp of
             loadn : (symtableentry : psym;symtable : psymtable);
             calln : (symtableprocentry : pprocsym;
                      symtableproc : psymtable;procdefinition : pprocdef;
                      methodpointer : ptree);
             ordconstn : (value : longint);
             realconstn : (valued : double;labnumber : longint);
             subscriptn : (vs : pvarsym);
             stringconstn : (values : pstring);
             typeconvn : (convtyp : tconverttype;explizit : boolean);
             inlinen : (inlinenumber : longint);
       end;

    procedure init_tree;
    function gennode(t : ttreetyp;l,r : ptree) : ptree;
    function genloadnode(v : pvarsym;st : psymtable) : ptree;
    function gensinglenode(t : ttreetyp;l : ptree) : ptree;
    function gensubscriptnode(varsym : pvarsym;l : ptree) : ptree;
    function genordinalconstnode(v : longint;def : pdef) : ptree;
    function gentypeconvnode(node : ptree;t : pdef) : ptree;
    function gencallparanode(expr,next : ptree) : ptree;
    function genrealconstnode(v : double) : ptree;
    function genasmnode : ptree;
    function gencallnode(v : pprocsym;st : psymtable) : ptree;
    function genmethodcallnode(v : pprocsym;st : psymtable;mp : ptree) : ptree;
    function genstringconstnode(const s : string) : ptree;
    function genzeronode(t : ttreetyp) : ptree;
    function geninlinenode(number : longint;l : ptree) : ptree;
    function gentypedconstloadnode(sym : ptypedconstsym;st : psymtable) : ptree;
    function genaufzaehlnode(v : paufzaehlsym) : ptree;
    function genselfnode(class : pdef) : ptree;

    function getcopy(p : ptree) : ptree;

    procedure disposetree(p : ptree);
    procedure putnode(p : ptree);
    function getnode : ptree;

    const
       flag_2_jmp : array[F_E..F_LE] of tasmop =
          (JE,JNE,JG,JL,JGE,JLE);

       flag_2_set : array[F_E..F_LE] of tasmop =
          (SETE,SETNE,SETG,SETL,SETGE,SETLE);

  implementation

    { eine Poolverwaltung fr die Nodes, was die Geschwindigkeit sehr }
    { steigert }

    var
       wurzel : ptree;

    procedure init_tree;

      begin
         wurzel:=nil;
      end;

    function getnode : ptree;

      var
         hp : ptree;

      begin
         if wurzel=nil then new(hp)
         else
           begin
              hp:=wurzel;
              wurzel:=wurzel^.left;
           end;

         { neue Node hat sicher keinen Fehler }
         hp^.error:=false;
         getnode:=hp;
      end;

    procedure putnode(p : ptree);

      begin
         p^.left:=wurzel;
         wurzel:=p;
      end;

    function getcopy(p : ptree) : ptree;

      var
         hp : ptree;

      begin
         hp:=getnode;
         hp^:=p^;

         { sehr unsauber: }
         if (p^.disposetyp=dt_leftright) or (p^.disposetyp=dt_left) then
           if assigned(hp^.left) then
             hp^.left:=getcopy(hp^.left);
         if (p^.disposetyp=dt_leftright) then
           if assigned(hp^.right) then
             hp^.right:=getcopy(hp^.right);

         if hp^.treetype=asmn then
           internalerror(11);
         getcopy:=hp;
      end;

    procedure disposetree(p : ptree);

      begin
         case p^.disposetyp of
            dt_leftright : begin
                              disposetree(p^.left);
                              disposetree(p^.right);
                           end;
            dt_nothing : ;
            dt_left    : begin
                            disposetree(p^.left);
                         end;
            dt_leftmbright    : begin
                            disposetree(p^.left);
                            if assigned(p^.right) then disposetree(p^.right);
                         end;
            dt_mbleft : if assigned(p^.left) then disposetree(p^.left);
            dt_mbleft_and_method : begin
                                      if assigned(p^.left) then disposetree(p^.left);
                                      disposetree(p^.methodpointer);
                                   end;
            dt_ltree : disposetree(p^.left);
            dt_string : stringdispose(p^.values);
            dt_typeconv : disposetree(p^.left);
            dt_inlinen : if assigned(p^.left) then
                           disposetree(p^.left);
            else internalerror(12);
         end;
         putnode(p);
      end;

    function gencallparanode(expr,next : ptree) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_leftmbright;
         p^.treetype:=callparan;
         p^.left:=expr;
         p^.right:=next;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=nil;
         gencallparanode:=p;
      end;

    function genasmnode : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_nothing;
         p^.treetype:=asmn;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=nil;
         genasmnode:=p;
      end;

    function gennode(t : ttreetyp;l,r : ptree) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_leftright;
         p^.treetype:=t;
         p^.left:=l;
         p^.right:=r;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=nil;
         gennode:=p;
      end;

    function genordinalconstnode(v : longint;def : pdef) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_nothing;
         p^.treetype:=ordconstn;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=def;
         p^.value:=v;
         genordinalconstnode:=p;
      end;

    function genaufzaehlnode(v : paufzaehlsym) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_nothing;
         p^.treetype:=ordconstn;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=v^.definition;
         p^.value:=v^.value;
         genaufzaehlnode:=p;
      end;

    function genrealconstnode(v : double) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_nothing;
         p^.treetype:=realconstn;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=cs64realdef;
         p^.valued:=v;
         p^.labnumber:=-1;
         genrealconstnode:=p;
      end;

    function genstringconstnode(const s : string) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_string;
         p^.treetype:=stringconstn;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=cstringdef;
         p^.values:=stringdup(s);
         genstringconstnode:=p;
      end;

    function gensinglenode(t : ttreetyp;l : ptree) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_left;
         p^.treetype:=t;
         p^.left:=l;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=nil;
         gensinglenode:=p;
      end;

    function genloadnode(v : pvarsym;st : psymtable) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.treetype:=loadn;
         p^.resulttype:=v^.definition;
         p^.symtableentry:=v;
         p^.symtable:=st;
         p^.disposetyp:=dt_nothing;
         genloadnode:=p;
      end;

    function gentypedconstloadnode(sym : ptypedconstsym;st : psymtable) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.treetype:=loadn;
         p^.resulttype:=sym^.definition;
         p^.symtableentry:=pvarsym(sym);
         p^.symtable:=st;
         p^.disposetyp:=dt_nothing;
         gentypedconstloadnode:=p;
      end;

    function gentypeconvnode(node : ptree;t : pdef) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_typeconv;
         p^.treetype:=typeconvn;
         p^.left:=node;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=t;
         p^.explizit:=false;
         gentypeconvnode:=p;
      end;

    function gencallnode(v : pprocsym;st : psymtable) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.treetype:=calln;

         p^.symtableprocentry:=v;
         p^.symtableproc:=st;
         p^.disposetyp:=dt_mbleft;
         p^.methodpointer:=nil;
         p^.left:=nil;
         p^.procdefinition:=nil;
         gencallnode:=p;
      end;

    function genmethodcallnode(v : pprocsym;st : psymtable;mp : ptree) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.treetype:=calln;

         p^.symtableprocentry:=v;
         p^.symtableproc:=st;
         p^.disposetyp:=dt_mbleft_and_method;
         p^.left:=nil;
         p^.methodpointer:=mp;
         p^.procdefinition:=nil;
         genmethodcallnode:=p;
      end;

    function gensubscriptnode(varsym : pvarsym;l : ptree) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_ltree;
         p^.treetype:=subscriptn;
         p^.left:=l;
         p^.registers32:=0;
         p^.vs:=varsym;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=nil;
         gensubscriptnode:=p;
      end;

   function genzeronode(t : ttreetyp) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_nothing;
         p^.treetype:=t;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=nil;
         genzeronode:=p;
      end;

    function genselfnode(class : pdef) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_nothing;
         p^.treetype:=selfn;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=class;
         genselfnode:=p;
      end;

    function geninlinenode(number : longint;l : ptree) : ptree;

      var
         p : ptree;

      begin
         p:=getnode;
         p^.disposetyp:=dt_inlinen;
         p^.treetype:=inlinen;
         p^.left:=l;
         p^.inlinenumber:=number;
         p^.registers32:=0;
{         p^.registers16:=0;
         p^.registers8:=0; }
         p^.registersfpu:=0;
         p^.resulttype:=nil;
         geninlinenode:=p;
      end;

end.
