{****************************************************************************

                   Copyright (c) 1993,94 by Florian Klmpfl

 ****************************************************************************}

unit types;

  interface

    uses
       cobjects,globals,asmgen,symtable,tree;

    { liefert true, wenn die bergebene Definition einen ordinalen Typ }
    { definiert                                                        }
    function is_ordinal(def : pdef) : boolean;

    { true, wenn beide Typen semantisch gleich sind }
    function is_equal(def1,def2 : pdef) : boolean;

    { gibt true zurueck, wenn zwei Parameterlisten gleich sind }
    function equal_paras(def1,def2 : pdefcoll) : boolean;

    { gibt den ordinalen Werten der Node zurueck oder falls sie }
    { keinen ordinalen Wert hat, wird ein Fehler erzeugt        }
    function get_ordinal_value(p : ptree) : longint;

    { prueft, ob l im gueltigen Bereich von def liegt }
    procedure testrange(def : pdef;l : longint);

    { erzeugt fr class die VMT }
    procedure genvmt(class : pclassdef);

    { true, falls p ein Zeiger auf eine konstanten Intknoten ist }
    function is_constintnode(p : ptree) : boolean;
    { entsprechend }
    function is_constboolnode(p : ptree) : boolean;
    function is_constrealnode(p : ptree) : boolean;
    function is_constcharnode(p : ptree) : boolean;

  implementation

    function is_constintnode(p : ptree) : boolean;

      begin
         is_constintnode:=((p^.treetype=ordconstn) and
           (p^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.resulttype)^.typ=s32bit));
      end;

    function is_constcharnode(p : ptree) : boolean;

      begin
         is_constcharnode:=((p^.treetype=ordconstn) and
           (p^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.resulttype)^.typ=uchar));
      end;

    function is_constrealnode(p : ptree) : boolean;

      begin
         is_constrealnode:=(p^.treetype=realconstn);
      end;

    function is_constboolnode(p : ptree) : boolean;

      begin
         is_constboolnode:=((p^.treetype=ordconstn) and
           (p^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.resulttype)^.typ=bool8bit));
      end;

    function equal_paras(def1,def2 : pdefcoll) : boolean;

      begin
         while (assigned(def1)) and (assigned(def2)) do
           begin
              if not(is_equal(def1^.data,def2^.data)) or
                 (def1^.paratyp<>def2^.paratyp) then
                begin
                   equal_paras:=false;
                   exit;
                end;
              def1:=def1^.next;
              def2:=def2^.next;
           end;
         if (def1=nil) and (def2=nil) then
           equal_paras:=true
         else
           equal_paras:=false;
      end;

    function is_ordinal(def : pdef) : boolean;

      var
         dt : tgrundtyp;

      begin
         is_ordinal:=false;
         if (def^.deftype<>grunddef) and (def^.deftype<>aufzaehldef) then
           exit;
         dt:=pgrunddef(def)^.typ;
         if ((def^.deftype=grunddef) and
           ((dt=s32bit) or
            (dt=uchar) or
            (dt=u8bit) or
            (dt=s8bit) or
            (dt=s16bit) or
            (dt=bool8bit) or
            (dt=u16bit))) or
           (def^.deftype=aufzaehldef) then is_ordinal:=true
      end;

    procedure testrange(def : pdef;l : longint);

      begin
         if def^.deftype=grunddef then
           case pgrunddef(def)^.typ of
              s32bit,s16bit,u16bit,s8bit,u8bit :
                if (l<pgrunddef(def)^.von) or (l>pgrunddef(def)^.bis) then
                  fatalerror(range_check_error);
              bool8bit : if (l<>0) or (l<>1) then
                           fatalerror(range_check_error);
              uchar : if (l<0) or (l>255) then
                           fatalerror(range_check_error);
           end
         else
           if def^.deftype=aufzaehldef then
             if (l<0) or (l>paufzaehldef(def)^.max) then
               fatalerror(range_check_error);
      end;

    function get_ordinal_value(p : ptree) : longint;

      begin
         if p^.treetype=ordconstn then
           get_ordinal_value:=p^.value
         else error(ordinal_expect);
      end;

    function is_equal(def1,def2 : pdef) : boolean;

      var
         b : boolean;
         hd : pdef;

      begin
         { Wenn ein String dabei, dann def1 zum String machen }
         if def2^.deftype=stringdef then
           begin
              hd:=def1;
              def1:=def2;
              def2:=hd;
           end;
         b:=false;

         { wenn beide auf die gleiche Definition zeigen sind sie wohl gleich...}
         if def1=def2 then
           b:=true
         else

           { wenn beides ein Pointertyp ist: }
           if (def1^.deftype=pointerdef) and (def2^.deftype=pointerdef) then
             begin
                { "void"-Pointer sind zu allen anderen Pointertypen kompatibel }
                if (ppointerdef(def1)^.definition^.deftype=grunddef) and
                   (pgrunddef(ppointerdef(def1)^.definition)^.typ=uvoid) then
                  b:=true
                else
                  if (ppointerdef(def2)^.definition^.deftype=grunddef) and
                     (pgrunddef(ppointerdef(def2)^.definition)^.typ=uvoid) then
                    b:=true
                else
                  { Pointer, die die gleiche Definition haben, sind auch gleich }
                  b:=is_equal(ppointerdef(def1)^.definition,ppointerdef(def2)^.definition)
             end
         else

         { Grundtypen sind gleich, wenn sie den selben Grundtyp haben, }
         { und wenn noetig den selben Unterbereich haben }
           if (def1^.deftype=grunddef) and (def2^.deftype=grunddef) then
             begin
                case pgrunddef(def1)^.typ of
                   u8bit,s32bit,s8bit,u16bit,s16bit : begin
                                     if pgrunddef(def1)^.typ=pgrunddef(def2)^.typ then
                                       if (pgrunddef(def1)^.von=pgrunddef(def2)^.von) and
                                          (pgrunddef(def1)^.bis=pgrunddef(def2)^.bis) then
                                           b:=true;
                                  end;
                   uvoid,s64real,bool8bit,uchar : b:=pgrunddef(def1)^.typ=pgrunddef(def2)^.typ;
                end;
             end
         else

            { Strings mit gleicher Laenge sind auch equivalent }
            if (def1^.deftype=stringdef) and (def2^.deftype=stringdef) and
               (pstringdef(def1)^.len=pstringdef(def2)^.len) then
            b:=true

(*          { STRING[N] ist equivalent zu ARRAY[0..N] OF CHAR (N<256) }
         else if ((def1^.deftype=stringdef) and (def2^.deftype=arraydef)) and
              (parraydef(def2)^.definition^.deftype=grunddef) and
              (pgrunddef(parraydef(def1)^.definition)^.typ=uchar) and
              (parraydef(def2)^.lowrange=0) and
              (parraydef(def2)^.highrange=pstringdef(def1)^.len) then
              b:=true *)
          else
            if (def1^.deftype=formaldef) and (def2^.deftype=formaldef) then
            b:=true;
         is_equal:=b;
      end;

    type
       pprocdefcoll = ^tprocdefcoll;

       tprocdefcoll = record
          next : pprocdefcoll;
          data : pprocdef;
       end;

       psymcoll = ^tsymcoll;

       tsymcoll = record
          next : psymcoll;
          name : pstring;
          data : pprocdefcoll;
       end;

    var
       wurzel : psymcoll;
       nextvirtnumber : longint;

    procedure eachsym(sym : psym);far;

      var
         procdefcoll : pprocdefcoll;
         hp : pprocdef;
         symcoll : psymcoll;
         _name : string;
         stored : boolean;

      begin
         { nur Unterprogrammsymbole werden in die VMT aufgenommen }
         if sym^.typ=procsym then
           begin
              _name:=sym^.name;
              symcoll:=wurzel;
              while assigned(symcoll) do
                begin
                   { wenn das Symbol in der Liste schon existiert }
                   if _name=symcoll^.name^ then
                     begin
                        { alle Definitionen des Symbols durchgehen }
                        hp:=pprocsym(sym)^.definition;
                        while assigned(hp) do
                          begin
                             { mit allen schon gespeicherten Definitionen }
                             { vergleichen                                }
                             procdefcoll:=symcoll^.data;
                             stored:=false;
                             while assigned(procdefcoll) do
                               begin
                                  { Parameter vergleichen }
                                  if equal_paras(procdefcoll^.data^.para1,hp^.para1) and
                                     (
                                       ((procdefcoll^.data^.options and povirtualmethod)<>0) or
                                       ((hp^.options and povirtualmethod)<>0)
                                     ) then
                                    begin
                                       { wenn sie gleich sind }
                                       { und eine davon virtual deklariert ist }
                                       { Fehler falls nur eine VIRTUAL }
                                       if (procdefcoll^.data^.options and povirtualmethod)<>
                                          (hp^.options and povirtualmethod) then
                                         error(overloaded_are_not_both_virtual);

                                       { Fehler falls Returntyp nicht gleich }
                                       if not(is_equal(procdefcoll^.data^.retdef,hp^.retdef)) then
                                         error(ol_meths_not_same_ret);

                                       { Fehler falls Excections nicht bereinstimmen }
                                       if (procdefcoll^.data^.options and poexceptions)<>
                                          (hp^.options and poexceptions) then
                                         error(overloaded_support_exceptions_false);

                                       { auch alle anderen Flags mssen bereinstimmen }
                                       if (procdefcoll^.data^.options<>hp^.options) then
                                         error(header_dont_match);

                                       { nun Nummer setzten }
                                       hp^.extnumber:=procdefcoll^.data^.extnumber;
                                       { und Austauschen: }
                                       procdefcoll^.data:=hp;
                                       stored:=true;
                                    end;
                                  procdefcoll:=procdefcoll^.next;
                               end;
                             { falls nicht in der Liste gespeichtert, }
                             { dann neu Eintragen                     }
                             if not(stored) then
                               begin
                                  new(procdefcoll);
                                  procdefcoll^.data:=hp;
                                  procdefcoll^.next:=symcoll^.data;
                                  symcoll^.data:=procdefcoll;
                                  { numerieren, falls VIRTUAL }
                                  if (hp^.options and povirtualmethod)<>0 then
                                    begin
                                       { dann numerieren }
                                       hp^.extnumber:=nextvirtnumber;
                                       { und die Nummer erhhen }
                                       inc(nextvirtnumber);
                                    end;
                               end;
                             hp:=hp^.nextoverloaded;
                          end;
                        exit;
                     end;
                   symcoll:=symcoll^.next;
                end;
              { falls nicht, Symbolitem neu erzeugen }
              new(symcoll);
              symcoll^.name:=stringdup(sym^.name);
              symcoll^.next:=wurzel;
              symcoll^.data:=nil;
              wurzel:=symcoll;
              hp:=pprocsym(sym)^.definition;
              { und alle Definitionen einfgen }
              while assigned(hp) do
                begin
                   new(procdefcoll);
                   procdefcoll^.data:=hp;
                   procdefcoll^.next:=symcoll^.data;
                   symcoll^.data:=procdefcoll;
                   { sollte eine virtueller Methodenaufruf dabei sein }
                   if (hp^.options and povirtualmethod)<>0 then
                     begin
                        { dann numerieren }
                        hp^.extnumber:=nextvirtnumber;
                        { und die Nummer erhhen }
                        inc(nextvirtnumber);
                     end;
                   { nchstes Element }
                   hp:=hp^.nextoverloaded;
                end;
           end;
      end;

    procedure genvmt(class : pclassdef);

      procedure do_genvmt(p : pclassdef);

        begin
           { bei der "ltesten Klasse anfangen }
           if assigned(p^.childof) then
             do_genvmt(p^.childof);
           { alle ffentlichen Symbole bearbeiten }
           p^.publicsyms^.foreach(eachsym);
        end;

      var
         symcoll : psymcoll;
         procdefcoll : pprocdefcoll;
         i : longint;

      begin
         wurzel:=nil;
         nextvirtnumber:=0;
         { "Gestrpp" fr VMT erzeugen }
         do_genvmt(class);
         { nun VMT mit brute force schreiben }
         { alle Nummern durchgehen }
         for i:=0 to nextvirtnumber-1 do
           begin
              symcoll:=wurzel;
              { alle Symbole durchgehen }
              while assigned(symcoll) do
                begin
                   { und alle berladenen Methoden durchgehen }
                   procdefcoll:=symcoll^.data;
                   while assigned(procdefcoll) do
                     begin
                        { Adresse in die VMT schreiben: }
                        { (natrliche nur die, die als VIRTUAL deklariert sind) }
                        if procdefcoll^.data^.extnumber=i then
                          if (procdefcoll^.data^.options and povirtualmethod)<>0 then
                            vmtasmlist.concat(gennasmrec(A_LONG,S_NO,
                              procdefcoll^.data^.mangledname));
                        procdefcoll:=procdefcoll^.next;
                     end;
                   symcoll:=symcoll^.next;
                end;
           end;
         { nun noch das "Getrpp" entfernen }
         symcoll:=wurzel;
         while assigned(symcoll) do
           begin
              wurzel:=symcoll^.next;
              stringdispose(symcoll^.name);
              procdefcoll:=symcoll^.data;
              while assigned(procdefcoll) do
                begin
                   symcoll^.data:=procdefcoll^.next;
                   dispose(procdefcoll);
                   procdefcoll:=symcoll^.data;
                end;
              dispose(symcoll);
              symcoll:=wurzel;
           end;
      end;

end.
